package paasio

const testVersion = 3

func NewWriteCounter() WriteCounter {
	w := new(WriteCounter)
	return w
}

func (w *WriteCounter) WriteCount() (int64, int) {
	return 0, 0
}

func NewReadCounter() ReadCounter {
	r := new(ReadCounter)
	return r
}

func (r *ReadCounter) ReadCount() (int64, int) {
	return 0, 0
}

func NewReadWriteCounter() *ReadWriteCounter {
	r := new(ReadWriteCounter)
	return r
}
